from math import dist
# from os import environ
from os.path import exists, isdir, isfile
from shutil import rmtree

from pytest import approx, mark

from fixtures import fixture, np, egAtomsXYZ, egAtomsNeighIdxs, egAtomsSurfIdxs, egAtomsWithSurfNeighIdxs
from sphractal.datasets import getExampleDataPath, getStrongScalingDataPath, getWeakScalingDataPaths, \
    getValidationDataPath, getCaseStudyDataPaths, getMiscellaneousDataPaths
from sphractal.utils import estDuration, getMinMaxXYZ, readInp, findNN, findSurf, calcDist, closestSurfAtoms, \
    oppositeInnerAtoms
from sphractal.surfVoxel import fibonacciSphere, pointsOnAtom, pointsToVoxels, voxelBoxCnts
from sphractal.surfExact import getNearFarCoord, scanBox, writeBoxCoords, findAtomsWithSurfNeighs, exactBoxCnts
from sphractal.boxCnt import voxelBoxCnts, exactBoxCnts, findSlope, runBoxCnt


EG_XYZ_ATOM_NUM = 670
ATOM_RAD = 1.69
MAX_RANGE = 36.58499999999998


@fixture
def egMinMaxXYZ():
    return np.array([404.065, 404.065, 404.065]), np.array([440.65, 440.65, 440.65])


@fixture
def egAtomsEle():
    return np.array(['Pd']*EG_XYZ_ATOM_NUM, dtype='U2')


@fixture
def egAtomsRad():
    return np.array([ATOM_RAD]*EG_XYZ_ATOM_NUM)


@fixture
def egVoxelBoxCnts():
    return ([-2.70926996, -2.40823997, -2.10720997, -1.80617997, -1.50514998, -1.20411998, -0.90308999, -0.60205999, -0.30103],
            [0.90308999, 1.50514998, 2.30963017, 2.91855453, 3.49192171, 4.05419158, 4.47568571, 4.51703746, 4.52071928])


@fixture
def egExactBoxCnts():
    return ([-0.23688234, -0.16309612, -0.10004438, -0.03477764,  0.03932408, 0.10260591,  0.17060299,  0.24023892,  0.30488175,  0.37314659],
            [3.20194306, 3.32139128, 3.5171959 , 3.68142216, 3.80522891, 3.9531796 , 4.09272064, 4.27207379, 4.38937875, 4.52953301])


@fixture
def egVoxelBoxCntDims():
    return 0.99646974, 2.11889159, np.array((1.94381059, 2.29397261))


@fixture
def egExactBoxCntDims():
    return 0.99660096, 2.24426950, np.array((2.11334077, 2.37519824))


def test_estDuration(egMinMaxXYZ):
    distDuration = estDuration(dist)
    distance, duration = distDuration(egMinMaxXYZ[0], egMinMaxXYZ[1])
    assert distance == approx(63.36707879), 'Incorrect output'
    assert isinstance(duration, float), 'Incorrect duration data type'


def test_getExampleDataPath():
    """Unit test of getExampleDataPath()."""
    inpFilePathAct = getExampleDataPath()
    assert isinstance(inpFilePathAct, str), 'getExampleDataPath() did not return a string'
    assert exists(inpFilePathAct), 'exampleOT.xyz not found'
    assert isfile(inpFilePathAct), 'exampleOT.xyz is not a file'


def test_getStrongScalingDataPath():
    """Unit test of getStrongScalingDataPath()."""
    inpFilePathAct = getStrongScalingDataPath()
    assert isinstance(inpFilePathAct, str), 'getStrongScalingDataPath() did not return a string'
    assert exists(inpFilePathAct), 'strongScalingSP.xyz not found'
    assert isfile(inpFilePathAct), 'strongScalingSP.xyz is not a file'


def test_getWeakScalingDataPaths():
    """Unit test of getWeakScalingDataPaths()."""
    inpFilePathsAct = getWeakScalingDataPaths()
    assert isinstance(inpFilePathsAct, list), 'getWeakScalingDataPaths() did not return a list'
    assert isinstance(inpFilePathsAct[0], str), 'Path in list is not str'
    assert len(inpFilePathsAct) == 12, 'Incorrect number of paths returned'
    assert exists(inpFilePathsAct[0]), 'Path in list not found'
    assert isfile(inpFilePathsAct[0]), 'Path in list not a file'


def test_getValidationDataPath():
    """Unit test of getValidationDataPath()."""
    inpFilePathAct = getValidationDataPath()
    assert isinstance(inpFilePathAct, str), 'getValidationDataPath() did not return a string'
    assert exists(inpFilePathAct), 'singleAtom.xyz not found'
    assert isfile(inpFilePathAct), 'singleAtom.xyz is not a file'


def test_getCaseStudyDataPaths():
    """Unit test of getCaseStudyDataPaths()."""
    inpFilePathsAct = getCaseStudyDataPaths()
    assert isinstance(inpFilePathsAct, list), 'getCaseStudyDataPaths() did not return a list'
    assert isinstance(inpFilePathsAct[0], str), 'Path in list is not str'
    assert len(inpFilePathsAct) == 18, 'Incorrect number of paths returned'
    assert exists(inpFilePathsAct[0]), 'Path in list not found'
    assert isfile(inpFilePathsAct[0]), 'Path in list not a file'


def test_getMiscellaneousDataPaths():
    """Unit test of getMiscellaneousDataPaths()."""
    inpFilePathsAct = getMiscellaneousDataPaths()
    assert isinstance(inpFilePathsAct, list), 'getMiscellaneousDataPaths() did not return a list'
    assert isinstance(inpFilePathsAct[0], str), 'Path in list is not str'
    assert len(inpFilePathsAct) == 3, 'Incorrect number of paths returned'
    assert exists(inpFilePathsAct[0]), 'Path in list not found'
    assert isfile(inpFilePathsAct[0]), 'Path in list not a file'


@mark.parametrize('radType, atomRad', [('atomic', ATOM_RAD), ('metallic', 1.37)])
def test_readInp(radType, atomRad, egAtomsEle, egAtomsXYZ):
    """Unit test of readInp()."""
    atomsEleAct, atomsRadAct, atomsXYZAct, maxRangeAct, minXYZAct, maxXYZAct = readInp(getExampleDataPath(), radType)

    assert isinstance(atomsEleAct, np.ndarray), 'atomsEle not ndarray'
    assert isinstance(atomsRadAct, np.ndarray), 'atomsRad not ndarray'
    assert isinstance(atomsXYZAct, np.ndarray), 'atomsXYZ not ndarray'

    assert atomsEleAct.dtype == '<U2', 'Incorrect atomsEle data type'

    assert np.all(atomsEleAct == egAtomsEle), 'Incorrect atomsEle values'
    assert atomsRadAct == approx([atomRad]*EG_XYZ_ATOM_NUM), 'Incorrect atomsRad values'
    assert atomsXYZAct == approx(egAtomsXYZ), 'Incorrect atomsXYZ values'

    assert maxRangeAct == approx(36.58499999999998), 'Incorrect maxRange value'
    assert minXYZAct == approx([404.065, 404.065, 404.065]), 'Incorrect minXYZ values'
    assert maxXYZAct == approx([440.65, 440.65, 440.65]), 'Incorrect maxXYZ values'


@mark.parametrize('maxAtomRad, radMult, totNeighNumExp, avgAvgBL', [(ATOM_RAD, 1.2, 6840, 2.87438906), (1.37, 1.5, 9774, 3.21771150)])
def test_findNN(maxAtomRad, radMult, totNeighNumExp, avgAvgBL, egMinMaxXYZ, egAtomsXYZ, egAtomsNeighIdxs):
    """
    Unit test of findNN(). 
    calcBL=False was not tested as there's no extra code for this conditional branch. 

    TODO: Test with extreme radMult.
    """
    egMinXYZ, egMaxXYZ = egMinMaxXYZ
    atomsRad = np.array([maxAtomRad] * EG_XYZ_ATOM_NUM)
    atomsNeighIdxsAct, atomsAvgBondLensAct = findNN(atomsRad, egAtomsXYZ, egMinXYZ, egMaxXYZ, maxAtomRad, radMult, True)
    totNeighNumAct = sum([len(atomNeighIdxs[atomNeighIdxs > -1]) for atomNeighIdxs in atomsNeighIdxsAct])

    assert isinstance(atomsNeighIdxsAct, np.ndarray), 'atomsNeighIdxs not ndarray'
    assert isinstance(atomsAvgBondLensAct, np.ndarray), 'atomsAvgBondLens not ndarray'

    assert totNeighNumAct == totNeighNumExp, 'Incorrect atomsNeighIdxs values'
    assert atomsAvgBondLensAct.mean() == approx(avgAvgBL), 'Incorrect atomsAvgBondLens values'


@mark.parametrize('findSurfAlg, numSurfAtomsExp', [('alphaShape', 326), ('convexHull', 6), ('numNeigh', 326)])
def test_findSurfAlgs(findSurfAlg, numSurfAtomsExp, egAtomsXYZ, egAtomsNeighIdxs):
    """
    Unit test of findSurf() outputs for different 'findSurfAlg' options.

    TODO: Test with extreme alpha.
    """
    atomsSurfIdxsAct = findSurf(egAtomsXYZ, egAtomsNeighIdxs, findSurfAlg, alpha=2.0 * ATOM_RAD)
    assert isinstance(atomsSurfIdxsAct, np.ndarray), 'atomsSurfIdxs not ndarray'
    assert len(atomsSurfIdxsAct) == numSurfAtomsExp, 'Incorrect number of surface atoms'


def test_findSurfAcc(egAtomsXYZ, egAtomsNeighIdxs, egAtomsSurfIdxs):
    """Unit test of findSurf() outputs accuracy."""
    atomsSurfIdxsAct = findSurf(egAtomsXYZ, egAtomsNeighIdxs, alpha=2.0 * ATOM_RAD)
    assert np.all(atomsSurfIdxsAct == egAtomsSurfIdxs), 'Incorrect surface atom indices'


def test_calcDist(egAtomsXYZ):
    """Unit test of calcDist()."""
    numTests = 100
    distsAct, distsExp = np.empty((numTests, numTests)), np.empty((numTests, numTests))
    for (i, p1) in enumerate(egAtomsXYZ):
        if i >= 100:
            continue
        for (j, p2) in enumerate(egAtomsXYZ):
            if j >= 100:
                continue
            distsAct[i][j] = calcDist(p1, p2)
            distsExp[i][j] = dist(p1, p2)
    assert distsAct == approx(distsExp), 'Incorrect calculated distance'


#@mark.parametrize('pointXYZ, surfNeighIdxs, idxPairExp', [(), (), ()])
#def test_closestSurfAtoms(pointXYZ, idxPairExp, egAtomsXYZ, egAtomsNeighIdxs):
#    """Unit test of closestSurfAtoms(). (To be implemented)"""
#    idxPairAct = closestSurfAtoms(pointXYZ, surfNeighIdxs, egAtomsXYZ, egAtomsNeighIdxs)
#    assert idxPairAct == idxPairExp, 'Incorrect atom indices pairs'


#@mark.parametrize('pointXYZ, atom1XYZ, atomNeighIdxs, isOppExp', [(), (), ()])
#def test_oppositeInnerAtoms(pointXYZ, atom1XYZ, atomNeighIdxs, isOppExp, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs):
#    """Unit test of oppositeInnerAtoms(). (To be implemented)"""
#    isOppAct = oppositeInnerAtoms(pointXYZ, atom1XYZ, atomNeighIdxs, 
#                                  egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs)
#    assert isOppAct == isOppExp, 'Incorrect results'


@mark.parametrize('numPoints, sphereRad, xyzsExp',
    [(3, 1.0, np.array([[0., 1., 0.], [-0.73736888, 0., -0.67549029], [0., -1., 0.]])),
     (5, 1.5, np.array([[0., 1.5, 0.], [-0.95787027, 0.75, -0.87748763], [0.13113859, 0., 1.49425656], [0.79038527, -0.75, -1.03091762], [-0., -1.5, 0.]])),
     (2, 100, np.array([[0., 100., 0.], [-0., -100., -0.]]))])
def test_fibonacciSphere(numPoints, sphereRad, xyzsExp):
    """Unit test of fibonacciSphere()."""
    xyzsAct = fibonacciSphere(numPoints, sphereRad)
    assert isinstance(xyzsAct, np.ndarray), 'Incorrect output data type'
    assert xyzsAct == approx(xyzsExp), 'Incorrect surface point coordinates'


#@mark.parametrize('atomIdx, numOuterPointsExp, numInnerPointsExp, rmInSurf', [(), (), ()])
#def test_pointsOnAtom(atomIdx, numOuterPointsExp, numInnerPointsExp, rmInSurf, egAtomsSurfIdxs, egAtomsRad, egAtomsXYZ, egAtomsNeighIdxs):
#    """Unit test of pointsOnAtom(). (To be implemented)"""
#    outerSurfsAct, innerSurfsAct = pointsOnAtom(atomIdx, 300, egAtomsSurfIdxs, egAtomsRad, egAtomsXYZ, egAtomsNeighIdxs, None, rmInSurf)
#    assert len(outerSurfsAct) == numOuterPointsExp, 'Incorrect number of outer surface points'
#    assert len(innerSurfsAct) == numInnerPointsExp, 'Incorrect number of inner surface points'


#@mark.parametrize('gridSize, numVoxelsExp', [(1024, 33204), (512, 33168), (256, 32888), (128, 29901), (64, 11329), (32, 3104)])
#def test_pointsToVoxels(gridSize, numVoxelsExp, egPointXYZs):
#    """Unit test of pointsToVoxels(). (To be implemented)"""
#    voxelXYZs, voxelIdxs = pointsToVoxels(egPointXYZs, gridSize)
#    assert len(voxelIdxs) == numVoxelsExp, 'Incorrect number of voxels'


#@mark.parametrize('scanBoxIdx, atomCoord', [(), (), ()])
#def test_getNearFarCoord(scanBoxIdx, atomCoord, scanBoxNearFarExp):
#    """Unit test of getNearFarCoord(). (To be implemented)"""
#    scanBoxNearFarAct = getNearFarCoord(scanBoxIdx, scanBoxLen, lowBound, atomCoord)
#    assert scanBoxNearFarAct == approx(scanBoxNearFarExp), 'Incorrect coordinates' 


#def test_scanBox():
#    """Unit test of scanBox(). (To be implemented)"""
#    belongTest = scanBox(minXYZ, scanBoxIdxs, scanBoxNearFarXYZs, scanBoxLen,
#                 atomIdx, atomRad, atomXYZ, atomNeighIdxs,
#                 atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
#                 rmInSurf=True)


def test_findAtomsWithSurfNeighs(egAtomsNeighIdxs, egAtomsSurfIdxs, egAtomsWithSurfNeighIdxs):
    """Unit test of findAtomsWithSurfNeighs()."""
    atomsWithSurfNeighIdxsAct = findAtomsWithSurfNeighs(egAtomsNeighIdxs, egAtomsSurfIdxs)
    assert np.all(atomsWithSurfNeighIdxsAct == egAtomsWithSurfNeighIdxs), 'Incorrect atom indices'


@mark.parametrize('trimLen', [True, False])
@mark.parametrize('visReg', [True, False])
@mark.parametrize('saveFig, figExists', [(True, True), (False, False)])
def test_findSlopeVis(trimLen, visReg, saveFig, figExists, egExactBoxCnts):
    """Unit test of findSlope() functionalities for different 'trimLen', 'visReg', and 'saveFig' options."""
    boxCntDimsAct = findSlope(egExactBoxCnts[0], egExactBoxCnts[1], 'example_ES', 'tests/outputs', trimLen, visReg=visReg, saveFig=saveFig)
    assert exists('./tests/outputs/boxCntDims/example_ES_boxCntDim.png') == figExists, 'example_ES_boxCntDim.png is not found'
    assert isfile('./tests/outputs/boxCntDims/example_ES_boxCntDim.png') == figExists, 'example_ES_boxCntDim.png is not a file'
    if isdir('./tests/outputs'):
        rmtree('./tests/outputs')


@mark.filterwarnings('ignore')
@mark.parametrize('minSample, isFinite1', [(-99, True), (11, False)])
@mark.parametrize('confLvl, isFinite2', [(-101, False), (99.9, True), (100, False)])
def test_findSlopeConfInt(minSample, isFinite1, confLvl, isFinite2, egExactBoxCnts):
    """Unit test of findSlope() outputs for different 'minSample' and 'confLvl' values."""
    boxCntDimsAct = findSlope(egExactBoxCnts[0], egExactBoxCnts[1], minSample=minSample, confLvl=confLvl)
    assert np.all(np.isfinite(boxCntDimsAct[2])) == (isFinite1 and isFinite2), 'Incorrect confidence interval'


@mark.parametrize('scales, counts, boxCntDimsExp', [([-2.70926996, -2.40823997, -2.10720997, -1.80617997, -1.50514998, -1.20411998, -0.90308999, -0.60205999, -0.30103], [0.90308999, 1.50514998, 2.30963017, 2.91855453, 3.49192171, 4.05419158, 4.47568571, 4.51703746, 4.52071928], (0.99646974, 2.11889159, np.array((1.94381059, 2.29397261)))), ([-0.23688234, -0.16309612, -0.10004438, -0.03477764, 0.03932408, 0.10260591, 0.17060299, 0.24023892, 0.30488175, 0.37314659], [3.20194306, 3.32139128, 3.5171959 , 3.68142216, 3.80522891, 3.9531796 , 4.09272064, 4.27207379, 4.38937875, 4.52953301], (0.99660096, 2.24426950, np.array((2.11334077, 2.37519824))))])
def test_findSlopeAcc(scales, counts, boxCntDimsExp):
    """Unit test of findSlope() outputs accuracy."""
    boxCntDimsAct = findSlope(scales, counts)
    assert isinstance(boxCntDimsAct[2], np.ndarray), 'Confidence interval returned is not ndarray'
    assert len(boxCntDimsAct[2]) == 2, 'Confidence interval returned is not two numbers'
    assert np.all(boxCntDimsAct[2] == approx(boxCntDimsExp[2])), 'Incorrect estimation of box-counting dimensions by findSlope()'
    assert boxCntDimsAct[:2] == approx(boxCntDimsExp[:2]), 'Incorrect estimation of box-counting dimensions by findSlope()'


#def test_getVoxelBoxCntVis(egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs):
#    """Unit test of voxelBoxCnts() functionalities to generate output files for visualisation (To be uncommented when compiled C++ code could be shipped together)."""
#    assert isfile(os.environ['FASTBC_EXE']), 'Executable not found at FASTBC_EXE'
#    voxelScalesAct, voxelCountsAct = voxelBoxCnts(egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs,
#                                                     'example', 'tests/outputs', verbose=True, genPCD=True)
#    assert exists('./tests/outputs/surfVoxelIdxs.txt'), 'surfVoxelBoxIdxs.txt is not found'
#    assert isfile('./tests/outputs/surfVoxelIdxs.txt'), 'surfVoxelBoxIdxs.txt is a file'
#    assert exists('./tests/outputs/surfVoxelBoxCnts.txt'), 'surfVoxelBoxCnts is not found'
#    assert isfile('./tests/outputs/surfVoxelBoxCnts.txt'), 'surfVoxelBoxCnts is not a file'
#    assert exists('./tests/outputs/surfPoints/example_surfPoints.xyz'), 'example_surfPoints.xyz is not found'
#    assert isfile('./tests/outputs/surfPoints/example_surfPoints.xyz'), 'example_surfPoints.xyz is not a file'
#    assert exists('./tests/outputs/surfPoints/example_surfPoints.pcd'), 'example_surfPoints.pcd is not found'
#    assert isfile('./tests/outputs/surfPoints/example_surfPoints.pcd'), 'example_surfPoints.pcd is not a file'
#    assert exists('./tests/outputs/surfVoxels/example_surfVoxels.xyz'), 'example_surfVoxels.xyz is not found'
#    assert isfile('./tests/outputs/surfVoxels/example_surfVoxels.xyz'), 'example_surfVoxels.xyz is not a file'
#    if isdir('./tests/outputs'):
#        rmtree('./tests/outputs')


#@mark.parametrize('rmInSurf, voxelScalesExp, voxelCountsExp', [(True, [-2.70926996, -2.40823997, -2.10720997, -1.80617997, -1.50514998, -1.20411998, -0.90308999, -0.60205999, -0.30103], [0.90308999, 1.50514998, 2.30963017, 2.91855453, 3.49192171, 4.05419158, 4.47568571, 4.51703746, 4.52071928]), (False, [-2.70926996, -2.40823997, -2.10720997, -1.80617997, -1.50514998, -1.20411998, -0.90308999, -0.60205999, -0.30103], [0.90308999, 1.50514998, 2.35024802, 3.02530587, 3.62479758, 4.14640714, 4.54740546, 4.5884958, 4.59508819])])
#def test_getVoxelBoxCntAcc(rmInSurf, voxelScalesExp, voxelCountsExp, egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs):
#    """Unit test of voxelBoxCnts() outputs accuracy (To be uncommented when compiled C++ code could be shipped together)."""
#    voxelScalesAct, voxelCountsAct = voxelBoxCnts(egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs,
#                                                     'example', 'tests/outputs', rmInSurf=rmInSurf, vis=False)
#    assert voxelScalesAct == approx(voxelScalesExp), 'Incorrect scales'
#    assert voxelCountsAct == approx(voxelCountsExp), 'Incorrect box counts'


@mark.parametrize('rmInSurf, exactScalesExp, exactCountsExp', [(True, [-0.23688234, -0.16309612, -0.10004438, -0.03477764, 0.03932408, 0.10260591, 0.17060299, 0.24023892, 0.30488175, 0.37314659], [3.20194306, 3.32139128, 3.5171959, 3.68142216, 3.80522891, 3.9531796, 4.09272064, 4.27207379, 4.38937875, 4.52953301]), (False, [-0.23688234, -0.16309612, -0.10004438, -0.03477764, 0.03932408, 0.10260591, 0.17060299, 0.24023892, 0.30488175, 0.37314659], [3.49789674, 3.69055046, 3.84997191, 4.00056421, 4.18301346, 4.32281862, 4.47290265, 4.61702132, 4.74061529, 4.86194043])])
def test_getExactBoxCnt(rmInSurf, exactScalesExp, exactCountsExp, egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs, egMinMaxXYZ):
    """Unit test of exactBoxCnts()."""
    exactScalesAct, exactCountsAct = exactBoxCnts(egAtomsEle, egAtomsRad, egAtomsSurfIdxs, egAtomsXYZ, egAtomsNeighIdxs,
                                                        MAX_RANGE, (ATOM_RAD*0.25, ATOM_RAD), egMinMaxXYZ[0], 'example', 'tests/outputs', rmInSurf=rmInSurf)
    print(exactScalesAct, exactCountsAct)
    assert exactCountsAct == approx(exactCountsExp), 'Incorrect scales'
    assert exactCountsAct == approx(exactCountsExp), 'Incorrect box counts'
    assert exists('./tests/outputs/boxCoords/example_boxCoords.xyz'), 'example_boxCoords.xyz is not found'
    assert isfile('./tests/outputs/boxCoords/example_boxCoords.xyz'), 'example_boxCoords.xyz is not a file'
    if isdir('./tests/outputs'):
        rmtree('./tests/outputs')


#def test_runBoxCnt(egVoxelBoxCntDims, egExactBoxCntDims):
#    """Unit and regression test of runBoxCnt() (To be uncommented when compiled C++ code could be shipped together)."""
#    assert isfile(environ['FASTBC_EXE']), 'Executable not found at FASTBC_EXE'
#    boxCntDimsAct = runBoxCnt(getExampleDataPath(), writeFileDir='tests/outputs')
#    assert boxCntDimsAct[:2] == approx(egVoxelBoxCntDims[:2]), 'Incorrect R2 and D_Box for point clouds representation'
#    assert boxCntDimsAct[2] == approx(egVoxelBoxCntDims[2]), 'Incorrect confidence interval for point clouds representation'
#    assert boxCntDimsAct[-3:-1] == approx(egExactBoxCntDims[:2]), 'Incorrect R2 and D_Box for exact surface representation'
#    assert boxCntDimsAct[-1] == approx(egExactBoxCntDims[2]), 'Incorrect confidence interval for exact surface representation'
#    if isdir('./tests/outputs'):
#        rmtree('./tests/outputs')




from concurrent.futures import ProcessPoolExecutor as Pool
from math import ceil, cos, log10, pi, sin, sqrt
from os import mkdir, sched_getaffinity, system
from os.path import isdir

from numba import njit
import numpy as np

from sphractal.constants import ATOMIC_RAD_DICT, METALLIC_RAD_DICT
from sphractal.utils import calcDist, oppositeInnerAtoms
# from sphractal.utils import annotate


@njit(fastmath=True, cache=True)
def fibonacciSphere(numPoints, sphereRad):
    """Generate evenly spread points on the surface of a sphere with a specified radius."""
    xyzs = np.empty((numPoints, 3), dtype=np.float64)
    phi = pi * (sqrt(5)-1)  # Golden angle (radians)
    for i in range(numPoints):
        y = 1 - (i / float(numPoints-1))*2  # y \in [1, -1]
        radius = sqrt(1 - y*y)  # Radius at y
        theta = phi * i  # Golden angle increment
        x = cos(theta) * radius
        z = sin(theta) * radius
        xyzs[i] = (x * sphereRad, y * sphereRad, z * sphereRad)
    return xyzs


@njit(fastmath=True, cache=True)
def withinNeighRad(surfPointXYZ, atomNeighIdxs, atomsRad, atomsXYZ):
    """Return whether a surface point generated around a given atom falls within the radius of a neighbouring atom."""
    for neighIdx in atomNeighIdxs:
        if calcDist(atomsXYZ[neighIdx], surfPointXYZ) <= atomsRad[neighIdx]:
            return True
    return False


@njit(fastmath=True, cache=True)
def rmPoint(args):
    """Check whether a point falls on the outer surface."""
    surfPoint, atomXYZ, atomNeighIdxs, atomsRad, atomsXYZ, rmInSurf, atomsSurfIdxs, atomsNeighIdxs = args
    surfPointXYZ = surfPoint + atomXYZ
    if withinNeighRad(surfPointXYZ, atomNeighIdxs, atomsRad, atomsXYZ):
        return 'isWithinRad', surfPointXYZ
    if (not rmInSurf) or (rmInSurf and oppositeInnerAtoms(surfPointXYZ, atomXYZ,
                                                          atomNeighIdxs, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs)):
        return 'toExclude', surfPointXYZ
    return 'toInclude', surfPointXYZ


# @annotate('pointsOnAtom', color='cyan')
def pointsOnAtom(args):
    """Generate surface points around an atom and classify them as either inner or outer surface."""
    atomIdx, numPoints, atomsSurfIdxs, atomsRad, atomsXYZ, atomsNeighIdxs, maxCPU, surfPoints, rmInSurf = args
    if surfPoints is None:
        surfPoints = fibonacciSphere(numPoints, atomsRad[atomIdx])
    atomXYZ, atomNeighIdxsPadded = atomsXYZ[atomIdx], atomsNeighIdxs[atomIdx]
    atomNeighIdxs = atomNeighIdxsPadded[atomNeighIdxsPadded > -1]
    rmPointInp, outerSurfs, innerSurfs = [], [], []

    # Include points that fall on surface of interest
    for surfPoint in surfPoints:
        if maxCPU > 1:
            rmPointInp.append((surfPoint, atomXYZ, atomNeighIdxs, atomsRad, atomsXYZ, rmInSurf, atomsSurfIdxs, atomsNeighIdxs))
            continue
        pointPosition, surfPointXYZ = rmPoint((surfPoint, atomXYZ, atomNeighIdxs, atomsRad, atomsXYZ, rmInSurf, atomsSurfIdxs, atomsNeighIdxs))
        if pointPosition == 'toExclude':
            outerSurfs.append(surfPointXYZ)
        elif pointPosition == 'toInclude':
            innerSurfs.append(surfPointXYZ)

    # Parallel implementation of the point position assessment algorithm
    if maxCPU > 1:
        with Pool(max_workers=maxCPU) as pool:
            for rmPointResult in pool.map(rmPoint, rmPointInp, 
                                          chunksize=ceil(numPoints / maxCPU)):
                if rmPointResult[0] == 'toExclude':
                    outerSurfs.append(rmPointResult[1])
                elif rmPointResult[0] == 'toInclude':
                    innerSurfs.append(rmPointResult[1])

    return outerSurfs, innerSurfs


# @annotate('pointsToVoxels', color='magenta')
@njit(fastmath=True, cache=True)
def pointsToVoxels(pointXYZs, gridSize):
    """Turn coordinates of point clouds into coordinates of occupied voxels."""
    # Get segmenting points in 3D space
    rangeXYZ = np.array([pointXYZs[:, i].max() - pointXYZs[:, i].min() for i in range(3)])
    maxRange = max(rangeXYZ)
    minXYZs, maxXYZs, segmentXYZs = np.empty(3), np.empty(3), np.empty((3, gridSize + 1))
    for i in range(3):
        minXYZs[i] = pointXYZs[:, i].min() - (maxRange-rangeXYZ[i]) * 0.5
        maxXYZs[i] = pointXYZs[:, i].max() + (maxRange-rangeXYZ[i]) * 0.5
        segmentXYZs[i] = np.linspace(minXYZs[i], maxXYZs[i], num=gridSize+1)

    # Get the index of the segment that each point lies within
    voxelXs = np.clip(np.searchsorted(segmentXYZs[0], pointXYZs[:, 0]) - 1, 0, gridSize)
    voxelYs = np.clip(np.searchsorted(segmentXYZs[1], pointXYZs[:, 1]) - 1, 0, gridSize)
    voxelZs = np.clip(np.searchsorted(segmentXYZs[2], pointXYZs[:, 2]) - 1, 0, gridSize)

    # Get indices of occupied voxels after flattened
    voxelIdxs = set(voxelXs*gridSize*gridSize + voxelYs*gridSize + voxelZs)
    return np.column_stack((voxelXs, voxelYs, voxelZs)), voxelIdxs


# @annotate('writeSurfVoxelIdxs', color='yellow')
def writeSurfVoxelIdxs(outDir, voxelIdxs):
    """Generate a txt file required for 3D box-counting using C++ code written by Ruiz de Miras and Posadas."""
    with open(f"{outDir}/surfVoxelIdxs.txt", 'w') as f:
        for idx in voxelIdxs:
            f.write(f"{idx}\n")


# @annotate('writePCD', color='yellow')
def writePCD(outDir, npName, surfPointXYZs):
    """Generate a pcd file required for 3D box-counting using MATLAB code written by Kazuaki Iida."""
    surfPointsDir = f"{outDir}/surfPoints"
    if not isdir(surfPointsDir):
        if not isdir(outDir):
            mkdir(outDir)
        mkdir(surfPointsDir)
    with open(f"{surfPointsDir}/{npName}_surfPoints.pcd", 'w') as f:
        f.write('# .PCD v.7 - Point Cloud Data file format\nVERSION .7')
        f.write('\nFIELDS x y z\nSIZE 4 4 4\nTYPE F F F\nCOUNT 1 1 1\n')
        f.write(f"WIDTH {len(surfPointXYZs)}\nHEIGHT 1\nPOINTS {len(surfPointXYZs)}\nDATA ascii\n")
        for xyz in surfPointXYZs:
            f.write(f"{xyz[0]} {xyz[1]} {xyz[2]}\n")


# @annotate('writeSurfPoints', color='blue')
def writeSurfPoints(outDir, npName, atomsSurfIdxs, atomsXYZ, surfPointXYZs, nonSurfPointXYZs):
    """Generate an xyz file for visualisation of classified point clouds."""
    surfPointsDir = f"{outDir}/surfPoints"
    if not isdir(surfPointsDir):
        if not isdir(outDir):
            mkdir(outDir)
        mkdir(surfPointsDir)
    with open(f"{surfPointsDir}/{npName}_surfPoints.xyz", 'w') as f:
        f.write(f"{len(surfPointXYZs) + len(nonSurfPointXYZs) + len(atomsSurfIdxs)}\n\n")
        for (i, xyz) in enumerate(surfPointXYZs):
            f.write(f"OU {xyz[0]} {xyz[1]} {xyz[2]} {i}\n")
        for (i, xyz) in enumerate(nonSurfPointXYZs):
            f.write(f"IN {xyz[0]} {xyz[1]} {xyz[2]} {i}\n")
        for atomIdx in atomsSurfIdxs:
            xyz = atomsXYZ[atomIdx]
            f.write(f"SU {xyz[0]} {xyz[1]} {xyz[2]} {atomIdx}\n")


# @annotate('writeSurfVoxels', color='green')
def writeSurfVoxels(outDir, npName, surfVoxelXYZs):
    """Generate an xyz file useful for visualisation of computed surface voxels."""
    surfVoxelsDir = f"{outDir}/surfVoxels"
    if not isdir(surfVoxelsDir):
        if not isdir(outDir):
            mkdir(outDir)
        mkdir(surfVoxelsDir)
    with open(f"{surfVoxelsDir}/{npName}_surfVoxels.xyz", 'w') as f:
        f.write(f"{len(surfVoxelXYZs)}\n\n")
        for (i, xyz) in enumerate(surfVoxelXYZs):
            f.write(f"VX {xyz[0]} {xyz[1]} {xyz[2]} {i}\n")


# @annotate('getSurfPoints', color='cyan')
def genSurfPoints(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                  npName, outDir='outputs', numCPUs=None, 
                  radType='atomic', numPoints=10000, gridNum=1024,
                  rmInSurf=True, vis=False, verbose=False, genPCD=False):
    """Generate point clouds approximating the outer spherical surface formed by a set of atoms."""
    # Avoid repeating generation of surface points around atoms with the same radii
    radDict = ATOMIC_RAD_DICT if radType == 'atomic' else METALLIC_RAD_DICT
    surfPointsEles = {atomEle: fibonacciSphere(numPoints, radDict[atomEle]) for atomEle in set(atomsEle)}

    # Resource allocations for parallelisation, rooms are available for further optimisation
    if numCPUs is None: 
        numCPUs = len(sched_getaffinity(0))
    minAtomCPU = max(1, len(atomsSurfIdxs) // 25)
    maxPointCPUperAtom = ceil(numPoints / numPoints)  # ceil(numPoints / 25)
    if numCPUs > maxPointCPUperAtom * minAtomCPU:
        atomConcMaxCPU = numCPUs // maxPointCPUperAtom
        pointConcMaxCPU = maxPointCPUperAtom
    # elif numCPUs > minAtomCPU:
    #     atomConcMaxCPU = minAtomCPU
    #     pointConcMaxCPU = numCPUs // minAtomCPU
    else:
        atomConcMaxCPU, pointConcMaxCPU = numCPUs, 1
    if verbose:
        print(f"    Assessing points over:\n      {len(atomsSurfIdxs)} atoms using {atomConcMaxCPU} cpu(s)...\n"
              f"      {numPoints} points using {pointConcMaxCPU} cpu(s)...")

    # Generate point clouds and convert to voxels
    surfPointXYZs, nonSurfPointXYZs = [], []
    pointsOnAtomInp = []
    for atomIdx in atomsSurfIdxs:
        if atomConcMaxCPU > 1:  # Adjust back
            pointsOnAtomInp.append((atomIdx, numPoints, atomsSurfIdxs, atomsRad, atomsXYZ, atomsNeighIdxs, pointConcMaxCPU,
                                    surfPointsEles[atomsEle[atomIdx]], rmInSurf))
        else:
            outerSurfs, innerSurfs = pointsOnAtom((atomIdx, numPoints, atomsSurfIdxs, atomsRad, atomsXYZ, atomsNeighIdxs, pointConcMaxCPU, surfPointsEles[atomsEle[atomIdx]], rmInSurf))
            surfPointXYZs.extend(outerSurfs)
            nonSurfPointXYZs.extend(innerSurfs)

    # Parallel implementation of the points generation around atoms
    if atomConcMaxCPU > 1:
        with Pool(max_workers=atomConcMaxCPU) as pool:
            for pointsOnAtomResult in pool.map(pointsOnAtom, pointsOnAtomInp, 
                                               chunksize=ceil(len(atomsSurfIdxs) / atomConcMaxCPU)):
                surfPointXYZs.extend(pointsOnAtomResult[0])
                nonSurfPointXYZs.extend(pointsOnAtomResult[1])

    surfVoxelXYZs, surfVoxelIdxs = pointsToVoxels(np.array(surfPointXYZs), gridNum)

    # Generate output files
    writeSurfVoxelIdxs(outDir, surfVoxelIdxs)
    if verbose:
        print(f"    {len(surfPointXYZs)} surface points -> {len(surfVoxelIdxs)} voxels, # grids: {gridNum}")
    if genPCD:
        writePCD(outDir, npName, surfPointXYZs)
    if vis:
        writeSurfPoints(outDir, npName, atomsSurfIdxs, atomsXYZ, surfPointXYZs, nonSurfPointXYZs)
        writeSurfVoxels(outDir, npName, surfVoxelXYZs)


# @annotate('voxelBoxCnts', color='blue')
def voxelBoxCnts(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                 npName, outDir='outputs', numCPUs=None, fastbcPath='$FASTBC',
                 radType='atomic', numPoints=300, gridNum=1024,
                 rmInSurf=True, vis=True, verbose=False, genPCD=False):
    """
    Count the boxes that cover the outer surface of a set of overlapping spheres represented as point clouds for different box sizes, using 3D box-counting algorithm written by Ruiz de Miras et al. in C++. 

    IMPORTANT: Make sure the source code has been downloaded from https://github.com/Jon-Ting/fastBC and compiled on your machine. 'fastbcPath' should point to the right file if FASTBC is not set as an environment variable.
    
    Parameters
    ----------
    atomsEle : 1D ndarray of strs
        Element type of each atom.
    atomsRad : 1D ndarray of floats
        Radius of each atom.
    atomsSurfIdxs : 1D ndarray of ints
        Indices of surface atoms.
    atomsXYZ : 2D ndarray of floats
        Cartesian coordinates of each atom.
    atomsNeighIdxs : 2D ndarray of ints
        Neighbour atoms indices of each atom.
    npName : str
        Identifier of the measured object, which forms part of the output file name, ideally unique.
    outDir : str, optional
        Path to the directory to store the output files.
    numCPUs : int, optional
        Number of CPUs to be used for parallelisation of tasks.
    fastbcPath : str, optional
        Path to the compiled C++ file for box-counting.
    radType : {'atomic', 'metallic'}, optional
        Type of radii to use for the spheres.
    numPoints : int, optional
        Number of surface points to be generated around each atom.
    gridNum : int, optional
        Resolution of the 3D binary image.
    rmInSurf : bool, optional
        Whether to remove the surface points on the inner surface.
    vis : bool, optional
        Whether to generate output files for visualisation.
    verbose : bool
        Whether to display the details.
    genPCD : bool, optional
        Whether to generate pcd file for box-counting using MATLAB code written by Kazuaki Iida.
    
    Returns
    -------
    scales : list
        Box lengths.
    counts : list
        Number of boxes that cover the surface of interest, as represented by the voxels in the 3D binary image.

    Examples
    --------
    >>> eles, rads, xyzs, _, minxyz, maxxyz = readInp('example.xyz')
    >>> neighs, _ = findNN(rads, xyzs, minxyz, maxxyz, 1.2)
    >>> surfs = findSurf(xyzs, neighs, 'alphaShape', 5.0)
    >>> scalesPC, countsPC = voxelBoxCnts(eles, rads, surfs, xyzs, neighs, 'example')

    Notes
    -----
    The 3D binary image resolution (gridNum) is restricted to 1024 or lower. Details about maximum grid size and memory estimation could be found in 'test.cpp' documented by the authors (https://www.ugr.es/~demiras/fbc/).
    """
    if verbose:
        print(f"  Approximating the surface with {numPoints} points for each atom...")
    if not isdir(outDir):
        mkdir(outDir)

    genSurfPoints(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                  npName, outDir, numCPUs,
                  radType, numPoints, gridNum,
                  rmInSurf, vis, verbose, genPCD)
    system(f"{fastbcPath} {gridNum} {outDir}/surfVoxelIdxs.txt {outDir}/surfVoxelBoxCnts.txt")
    scales, counts = [], []
    with open(f"{outDir}/surfVoxelBoxCnts.txt", 'r') as f:
        for line in f:
            scales.append(log10(1 / int(line.split()[0])))
            counts.append(log10(int(line.split()[1])))
    return scales[::-1], counts[::-1]



import setuptools


__version__ = '1.1.0'
with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="sphractal",
    version=__version__,
    author="Jonathan Yik Chang Ting",
    author_email="jonting97@gmail.com",
    description="Package to estimate surface fractal dimension of 3D objects composed of overlapping spheres via box-counting algorithm.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jon-Ting/sphractal",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License"
    ),
)

